/*
 * io.c:
 * Various IO functions.
 *
 * Copyright (c) 2004 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

static const char rcsid[] = "$Id: io.c,v 1.8 2004/04/25 01:27:17 chris Exp $";

#include <sys/types.h>

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include "consolidate.h"

/* xwrite FD BUF LEN
 * Like write(2), but ignoring signals and making up for short writes. */
ssize_t xwrite(int fd, const char *buf, size_t len) {
    size_t l;
    l = len;
    while (l > 0) {
        ssize_t n;
        do
            n = write(fd, buf, l);
        while (n == -1 && errno == EINTR);
        if (n == -1)
            return -1;
        else {
            l -= n;
            buf += n;
        }
    }
    return len;
}

/* xread FD BUF LEN
 * Like read(2), but ignoring signals. */
ssize_t xread(int fd, char *buf, size_t len) {
    ssize_t n;
    do
        n = read(fd, buf, len);
    while (n == -1 && errno == EINTR);
    return n;
}

/* write_to_tty FD BUFFER LEN
 * Write LEN bytes from BUFFER to the TTY, returning 1 on success or 0 on
 * failure. */
int write_to_tty(int fd, const char *buf, const size_t len) {
    return xwrite(fd, buf, len) == len;
}

/* write_to_client TTY FD BUFFER LEN
 * Write LEN bytes from BUFFER to the FD, returning 1 on success or 0 on
 * failure. */
int write_to_client(const char *tty, int fd, const char *buf, const size_t len) {
    ssize_t n;
    n = xwrite(fd, buf, len);
    if (-1 == n) {
        if (errno == EAGAIN)
            log_print(LOG_ERR, "%s: connected client stalled", tty);
        else
            log_print(LOG_ERR, "%s: connected client: %m", tty);
        return 0;
    } else
        return 1;
}

/* write_to_logfile TTY NAME FD BUFFER LEN
 * Write LEN bytes from BUFFER to the FD, prepending a timestamp and escaping
 * non-printable characters. Returns the file descriptor of the log file (which
 * will change when the file is rotated) on success, or -1 on failure. */
size_t maxlogsize, maxlogfiles;
int write_to_logfile(const char *tty, const char *name, int fd, const char *line, size_t len) {
    static char *buf;
    static size_t buflen;
    time_t T;
    const char *p;
    char *q;
    off_t o;

    while (len > 0 && isspace(line[len - 1])) --len;
    if (len == 0) return fd;

    if (buflen < 18 + len * 4 + 1)
        buf = xrealloc(buf, buflen = 18 + len * 4 + 1);

    /* ISO format: yyyymmddThhmmss */
    time(&T);
    strftime(buf, buflen, "%Y%m%dT%H%M%S: ", localtime(&T));

    for (p = line, q = buf + 17; p < line + len; ++p) {
        if (*p != '\\' && isprint(*p) && (*p == ' ' || !isspace(*p)))
            *q++ = *p;
        else {
            switch (*p) {
                case '\n': q += sprintf(q, "\\n");
                           break;
                case '\r': q += sprintf(q, "\\r");
                           break;
                case '\t': q += sprintf(q, "\\t");
                           break;
                case '\\': q += sprintf(q, "\\\\");
                           break;
                default:   q += sprintf(q, "\\x%02x", ((unsigned)*p) & 0xff);
                           break;
            }
        }
    }

    *q++ = '\n';

    if (xwrite(fd, buf, q - buf) != (q - buf)) {
        log_print(LOG_ERR, "%s: %s: write: %m", tty, name);
        close(fd);
        return -1;
    } else if (-1 == (o = lseek(fd, 0, SEEK_CUR))) {
        close(fd);
        return -1;
    } else if (o > maxlogsize) {
        close(fd);
        if (!rotate_logfile(name, 1, maxlogfiles))
            log_print(LOG_ERR, "%s: %s: error rotating logs: %m", tty, name);
        if (-1 == (fd = open(name, O_WRONLY | O_CREAT | O_APPEND, 0666)))
            log_print(LOG_ERR, "%s: %s: reopening log file: %m", tty, name);
    }

    return fd;
}
