/*
 * logrotate.c:
 * Log file rotation.
 *
 * Copyright (c) 2004 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

static const char rcsid[] = "$Id: logrotate.c,v 1.3 2004/08/13 19:19:31 chris Exp $";

#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <sys/stat.h>

#include "consolidate.h"

/* file_exists FILE
 * Does the named FILE exist? */
static int file_exists(const char *file) {
    struct stat st;
    return stat(file, &st) == 0 || errno != ENOENT;
}

/* rotate_one_logfile LOGFILE NUMBER MAX
 * Rotate a single logfile with the given NUMBER into the next numbered one,
 * unless NUMBER is equal to or larger than MAX, in which case it will be
 * deleted. Returns true on success or false on failure. */
static int rotate_one_logfile(const char *logfile, const unsigned int n, const unsigned int max) {
    char *buf = NULL, *buf2 = NULL;
    int np1 = 0, ret = 0;
    
    buf = xmalloc(strlen(logfile) + 32);
    
    /* See whether there is an older logfile we need to move out of the way. */
    sprintf(buf, "%s.%u", logfile, n + 1);
    if (file_exists(buf))
        np1 = 1;
    else {
        sprintf(buf, "%s.%u.gz", logfile, n + 1);
        if (file_exists(buf))
            np1 = 1;
    }

    if (np1 && !rotate_one_logfile(logfile, n + 1, max))
        goto fail;

    buf2 = xmalloc(strlen(logfile) + 32);

    /* Move this log file, whether gzipped or not. */
    if (n)
        sprintf(buf, "%s.%u", logfile, n);
    else
        strcpy(buf, logfile);
    sprintf(buf2, "%s.%u", logfile, n + 1);
    if (max && n >= max) {
        if (unlink(buf) == -1 && errno != ENOENT)
            goto fail;
    } else if (rename(buf, buf2) == -1 && errno != ENOENT)
        goto fail;
    

    if (n) {
        /* No gzipped n == 0 logfile. */
        sprintf(buf, "%s.%u.gz", logfile, n);
        sprintf(buf2, "%s.%u.gz", logfile, n + 1);
        if (max && n >= max) {
            if (unlink(buf) == -1 && errno != ENOENT)
                goto fail;
        } else if (rename(buf, buf2) == -1 && errno != ENOENT)
            goto fail;
    }

    ret = 1;

fail:
    xfree(buf);
    xfree(buf2);
    return ret;
}

/* rotate_logfile LOGFILE GZIP MAX
 * Rotate the named LOGFILE, so that LOGFILE is renamed to LOGFILE.1, LOGFILE.1
 * to LOGFILE.2, and so forth. If GZIP is true, then some effort will be made
 * to compress old logfiles with gzip(1). If MAX is greater than zero, then it
 * indicates the largest number of old logfiles which will be kept. Returns
 * true on success or false on failure. */
int rotate_logfile(const char *logfile, const int gzip, const unsigned int max) {
    int ret = 0;
    if ((ret = rotate_one_logfile(logfile, 0, max)) && gzip) {
        /* Do this in the background. */
        pid_t p;
        if ((p = fork()) == 0) {
            char *cmd;
            cmd = sprintf_new("gzip %s.[0-9]*", logfile);
            system(cmd);
            xfree(cmd);
            _exit(0);
        } else if (p == -1)
            ret = 0;
        /* else we are the parent; let SIGCHLD handler clean it up. */
    }
    return ret;
}

