/*
 * tty.c:
 * Terminal stuff.
 *
 * Copyright (c) 2004 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

static const char rcsid[] = "$Id: tty.c,v 1.2 2004/06/22 19:51:00 chris Exp $";

#include <sys/types.h>

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <termios.h>
#include <unistd.h>

#include "consolidate.h"
#include "tty.h"

/* parse_serial_options STRING OPTIONS
 * Save in OPTIONS the serial options described by STRING, returning 1 on
 * success or 0 on failure. STRING is of the form RATE[,PARITY[,DATABITS]]. */
int parse_serial_options(const char *str, struct serial_options *S) {
    int n, m;
    char c;
    switch (sscanf(str, "%d,%c,%d", &n, &c, &m)) {
        case 3:
            if (m < 5 || m > 8)
                return 0;
            else
                S->so_databits = m;
            /* fall through */

        case 2:
            switch (tolower(c)) {
                case 'n':   S->so_parity = p_none;
                            break;
                case 'o':   S->so_parity = p_odd;
                            break;
                case 'e':   S->so_parity = p_even;
                            break;
                default:    return 0;
            }
            /* fall through */

        case 1:
            if (B0 == (S->so_rate = n_to_rate(n)))
                return 0;
            else
                return 1;

        case 0:
        default:
            return 0;
    }
}

/* describe_serial_options OPTIONS
 * Return a human-readable string describing OPTIONS. */
char *describe_serial_options(const struct serial_options *O) {
    static char buf[80];
    sprintf(buf, "%d bps, %d data bits, %s parity, %s flow control",
                rate_to_n(O->so_rate),
                O->so_databits,
                O->so_parity == p_none ? "no" :
                    (O->so_parity == p_even ? "even" : "odd"),
                O->so_flowcontrol == f_none ? "no" :
                    (O->so_flowcontrol == f_hw ? "hardware" : "software"));
    return buf;
}

/* get_serial_options FD OPTIONS
 * Save in *OPTIONS the serial options on the TTY connected to FD. Returns true
 * on success or false on failure. */
int get_serial_options(int fd, struct serial_options *O) {
    struct termios T;

    if (-1 == tcgetattr(fd, &T))
        return 0;

    O->so_rate = cfgetispeed(&T);

    if (T.c_cflag & CS8)
        O->so_databits = 8;
    else if (T.c_cflag & CS7)
        O->so_databits = 7;
    else if (T.c_cflag & CS6)
        O->so_databits = 6;
    else if (T.c_cflag & CS5)
        O->so_databits = 5;
    else
        abort(); /* NOTREACHED */

    O->so_parity = p_none;
    if (T.c_cflag & PARODD)
        O->so_parity = p_odd;
    else if (T.c_cflag & PARENB)
        O->so_parity = p_even;

    O->so_flowcontrol = f_none;
    if (T.c_cflag & CRTSCTS)
        O->so_flowcontrol = f_hw;
    else if (T.c_iflag & (IXON | IXOFF))
        O->so_flowcontrol = f_sw;
    
    return 1;
}


/* set_serial_options TTY OPTIONS
 * Apply to TTY the passed OPTIONS. Returns true on success or false on
 * failure. */
int set_serial_options(int fd, const struct serial_options *O) {
    struct termios T = {0};

    if (-1 == tcgetattr(fd, &T))
        return 0;
    
    cfsetispeed(&T, O->so_rate);
    cfsetospeed(&T, O->so_rate);
    
    /* APUE p354--355 */ /* cfmakeraw XXX ? */
    T.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
    T.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON | IXOFF);
    T.c_oflag &= ~(OPOST);
    T.c_cflag &= ~(CSIZE | PARENB | PARODD);

    switch (O->so_databits) {
        case 5:
            T.c_cflag |= CS5;
            break;
        case 6:
            T.c_cflag |= CS6;
            break;
        case 7:
            T.c_cflag |= CS7;
            break;
        case 8:
            T.c_cflag |= CS8;
            break;
        default:
            abort();    /* NOTREACHED */
    }

    switch (O->so_parity) {
        case p_none:
            break; /* no change needed */
        case p_odd:
            T.c_cflag |= PARODD;
            /* fall through */
        case p_even:
            T.c_cflag |= PARENB;
            break;
    }

    switch (O->so_flowcontrol) {
        case f_none:
            break;  /* no change needed */
        case f_hw:
            T.c_cflag |= CRTSCTS;
            break;
        case f_sw:
            T.c_iflag |= IXON | IXOFF;
            break;
    }

    if (-1 == tcsetattr(fd, TCSANOW, &T))
        return 0;
    else
        return 1;
}
