/*
 * uucplock.c:
 * Implementation of UUCP lock files for ttys.
 *
 * Copyright (c) 2004 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

static const char rcsid[] = "$Id: uucplock.c,v 1.1 2004/08/15 17:06:28 chris Exp $";

#include <sys/types.h>

#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "consolidate.h"

#ifndef UUCP_LOCK_DIR
#   warning UUCP_LOCK_DIR not defined; using default, "/var/lock"
#   define UUCP_LOCK_DIR    "/var/lock"
#endif

/* 
 * A tty, /dev/$name, is locked with the file UUCP_LOCK_DIR/LCK..$name. That
 * file should contain the PID of the process which created the lock, either
 * as an ASCII string or straight binary.
 */

/* lockfile_name DEVICE
 * Return the lockfile name which should be used for the named DEVICE, or
 * NULL on failure. */
static char *lockfile_name(const char *dev) {
    static char buf[PATH_MAX];
    const char *s;
    char *d;
    s = strchr(dev + 1, '/');
    if (!s) return NULL;
    ++s;
    d = buf + sprintf(buf, "%s/LCK..", UUCP_LOCK_DIR);
    /* "/" in device name is translated to "_". At least, that's what picocom
     * does, so why change? */
    while (*s) {
        if (*s == '/') {
            *d++ = '_';
            s++;
        } else
            *d++ = *s++;
    }
    *d = 0;
    return buf;
}

/* uucp_lock DEVICE
 * Lock DEVICE, returning true on success or false on failure. */
int uucp_lock(const char *dev) {
    char *l;
    int fd = -1, ret = 0;
    pid_t p = 0;
    char buf[32];

    l = lockfile_name(dev);

    if (-1 != (fd = open(l, O_RDONLY))) {
        ssize_t n;

        n = xread(fd, buf, -1 + sizeof buf);

        if (n > 0) {
            char *lf;

            buf[n] = 0;
            
            /* These data could be either a binary integer or an ASCII decimal
             * integer. */
            if ((lf = strchr(buf, '\n')) && lf > buf && lf == (buf + strspn(buf, " 0123456789")))
                /* Looks like ASCII. */
                p = (pid_t)atoi(buf + strspn(buf, " "));

            if (!p) {
                if (n == sizeof(pid_t))
                    memcpy(&p, buf, sizeof p);
                else if (n == sizeof(int)) {
                    int m;
                    memcpy(&m, buf, sizeof m);
                    p = (pid_t)m;
                }
            }

            if (p <= 1) {
                log_print(LOG_WARNING, "%s: lockfile, %s, is bogus; removing it", dev, l);
                unlink(l);
            } else {
                /* It could be a valid PID. */
                errno = 0;
                if (kill(p, 0) == -1 && errno == ESRCH) {
                    /* No such process; lock is stale. */
                    log_print(LOG_WARNING, "%s: lockfile, %s, is bogus; removing it", dev, l);
                    unlink(l);
                } else {
                    /* A process exists with the given PID. */
                    errno = EEXIST;
                    goto fail;
                }
            }
        }
    }

    if (-1 == (fd = open(l, O_WRONLY | O_CREAT | O_EXCL, 0666)))
        goto fail;

    sprintf(buf, "%d\n", (int)getpid());
    if (xwrite(fd, buf, strlen(buf)) != strlen(buf)) {
        unlink(l);
        goto fail;
    }

    ret = 1;
    /* XXX should we lock the file and hold it open now? */
    
fail:
    if (fd != -1) close(fd);
    return ret;
}

/* uucp_unlock DEVICE
 * Unlock DEVICE. Should be called only if the lock is held by this process. */
void uucp_unlock(const char *dev) {
    unlink(lockfile_name(dev));
}

